



if exist %PREFIX%\\Library\\include\\manif\\manif.h (exit 0) else (exit 1)
IF %ERRORLEVEL% NEQ 0 exit /B 1
if exist %PREFIX%\\Library\\share\\manif\\cmake\\manifConfig.cmake (exit 0) else (exit 1)
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
